/*
 * $Id: awk.c,v 1.101 2006-10-22 12:39:30 bacon Exp $
 */

#include <sse/awk/awk.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>

#ifdef SSE_CHAR_IS_WCHAR
	#include <wchar.h>
	#include <wctype.h>
#endif

#ifndef __STAND_ALONE
	#include <xp/bas/stdio.h>
	#include <xp/bas/stdlib.h>
	#include <xp/bas/string.h>
	#include <xp/bas/memory.h>
	#include <xp/bas/sysapi.h>
	#include <xp/bas/assert.h>	
	#include <xp/bas/locale.h>	
#else
	#include <xp/bas/stdio.h>
	#include <limits.h>
	#ifndef PATH_MAX
		#define SSE_PATH_MAX 4096
	#else
		#define SSE_PATH_MAX PATH_MAX
	#endif

	#ifdef _WIN32
		#define xp_sprintf _sntprintf
	#endif
#endif

#ifdef _WIN32
#include <windows.h>
#include <tchar.h>
#pragma warning (disable: 4996)
#endif

#ifdef __STAND_ALONE
	#include <assert.h>
	#define xp_assert assert
#endif

#if defined(_WIN32) && defined(_MSC_VER) && defined(_DEBUG)
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#endif

#if defined(__linux) && defined(_DEBUG)
#include <mcheck.h>
#endif

struct src_io
{
	const xp_char_t* input_file;
	FILE* input_handle;
};

static FILE* fopen_t (const xp_char_t* path, const xp_char_t* mode)
{
#ifdef _WIN32
	return _tfopen (path, mode);
#else
	#ifdef SSE_CHAR_IS_MCHAR
	const xp_mchar_t* path_mb;
	const xp_mchar_t* mode_mb;
	#else
	xp_mchar_t path_mb[XP_PATH_MAX + 1];
	xp_mchar_t mode_mb[32];
	#endif

	#ifdef SSE_CHAR_IS_MCHAR
	path_mb = path;
	mode_mb = mode;
	#else
	if (xp_wcstomcs_strict (
		path, path_mb, xp_countof(path_mb)) == -1) return SSE_NULL;
	if (xp_wcstomcs_strict (
		mode, mode_mb, xp_countof(mode_mb)) == -1) return SSE_NULL;
	#endif

	return fopen (path_mb, mode_mb);
#endif
}

static int __dprintf (const xp_char_t* fmt, ...)
{
	int n;
	va_list ap;
#ifdef _WIN32
	xp_char_t buf[1024];
#endif

	va_start (ap, fmt);
#ifdef _WIN32
	n = xp_vsprintf (buf, xp_countof(buf), fmt, ap);
#if defined(_MSC_VER) && (_MSC_VER>=1400)
	MessageBox (NULL, buf, SSE_T("\uD655\uC778\uC2E4\uD328 Assertion Failure"), MB_OK | MB_ICONERROR);
#else
	MessageBox (NULL, buf, SSE_T("Assertion Failure"), MB_OK | MB_ICONERROR);
#endif

#else
	n = xp_vprintf (fmt, ap);
#endif
	va_end (ap);
	return n;
}

static FILE* popen_t (const xp_char_t* cmd, const xp_char_t* mode)
{
#ifdef _WIN32
	return _tpopen (cmd, mode);
#else
	#ifdef SSE_CHAR_IS_MCHAR
	const xp_mchar_t* cmd_mb;
	const xp_mchar_t* mode_mb;
	#else
	xp_mchar_t cmd_mb[2048];
	xp_mchar_t mode_mb[32];
	#endif

	#ifdef SSE_CHAR_IS_MCHAR
	cmd_mb = cmd;
	mode_mb = mode;
	#else
	if (xp_wcstomcs_strict (
		cmd, cmd_mb, xp_countof(cmd_mb)) == -1) return SSE_NULL;
	if (xp_wcstomcs_strict (
		mode, mode_mb, xp_countof(mode_mb)) == -1) return SSE_NULL;
	#endif

	return popen (cmd_mb, mode_mb);
#endif
}

#ifdef WIN32
	#define fgets_t _fgetts
	#define fputs_t _fputts
#else
	#ifdef SSE_CHAR_IS_MCHAR
		#define fgets_t fgets
		#define fputs_t fputs
	#else
		#define fgets_t fgetws
		#define fputs_t fputws
	#endif
#endif

static xp_ssize_t process_source (
	int cmd, void* arg, xp_char_t* data, xp_size_t size)
{
	struct src_io* src_io = (struct src_io*)arg;
	xp_char_t c;

	if (cmd == SSE_AWK_IO_OPEN)
	{
		if (src_io->input_file == SSE_NULL) return 0;
		src_io->input_handle = fopen_t (src_io->input_file, SSE_T("r"));
		if (src_io->input_handle == NULL) return -1;
		return 1;
	}
	else if (cmd == SSE_AWK_IO_CLOSE)
	{
		if (src_io->input_file == SSE_NULL) return 0;
		fclose ((FILE*)src_io->input_handle);
		return 0;
	}
	else if (cmd == SSE_AWK_IO_READ)
	{
		if (size <= 0) return -1;
	#ifdef SSE_CHAR_IS_MCHAR
		c = fgetc ((FILE*)src_io->input_handle);
	#else
		c = fgetwc ((FILE*)src_io->input_handle);
	#endif
		if (c == SSE_CHAR_EOF) return 0;
		*data = c;
		return 1;
	}

	return -1;
}

static xp_ssize_t dump_source (
	int cmd, void* arg, xp_char_t* data, xp_size_t size)
{
	/*struct src_io* src_io = (struct src_io*)arg;*/

	if (cmd == SSE_AWK_IO_OPEN) return 1;
	else if (cmd == SSE_AWK_IO_CLOSE) return 0;
	else if (cmd == SSE_AWK_IO_WRITE)
	{
		xp_size_t i;
		for (i = 0; i < size; i++)
		{
	#ifdef SSE_CHAR_IS_MCHAR
			fputc (data[i], stdout);
	#else
			fputwc (data[i], stdout);
	#endif
		}
		return size;
	}

	return -1;
}

static xp_ssize_t process_extio_pipe (
	int cmd, void* arg, xp_char_t* data, xp_size_t size)
{
	sse_awk_extio_t* epa = (sse_awk_extio_t*)arg;

	switch (cmd)
	{
		case SSE_AWK_IO_OPEN:
		{
			FILE* handle;
			const xp_char_t* mode;

			if (epa->mode == SSE_AWK_IO_PIPE_READ)
				mode = SSE_T("r");
			else if (epa->mode == SSE_AWK_IO_PIPE_WRITE)
				mode = SSE_T("w");
			else return -1; /* TODO: any way to set the error number? */
xp_printf (SSE_TEXT("opending %s of type %d (pipe)\n"),  epa->name, epa->type);
			handle = popen_t (epa->name, mode);
			if (handle == NULL) return -1;
			epa->handle = (void*)handle;
			return 1;
		}

		case SSE_AWK_IO_CLOSE:
		{
xp_printf (SSE_TEXT("closing %s of type (pipe) %d\n"),  epa->name, epa->type);
			fclose ((FILE*)epa->handle);
			epa->handle = NULL;
			return 0;
		}

		case SSE_AWK_IO_READ:
		{
			if (fgets_t (data, size, (FILE*)epa->handle) == SSE_NULL) 
				return 0;
			return sse_awk_strlen(data);
		}

		case SSE_AWK_IO_WRITE:
		{
			/* TODO: size... */
			fputs_t (data, (FILE*)epa->handle);
			return size;
		}

		case SSE_AWK_IO_FLUSH:
		{
			if (epa->mode == SSE_AWK_IO_PIPE_READ) return -1;
			else return 0;
		}

		case SSE_AWK_IO_NEXT:
		{
			return -1;
		}
	}

	return -1;
}

static xp_ssize_t process_extio_file (
	int cmd, void* arg, xp_char_t* data, xp_size_t size)
{
	sse_awk_extio_t* epa = (sse_awk_extio_t*)arg;

	switch (cmd)
	{
		case SSE_AWK_IO_OPEN:
		{
			FILE* handle;
			const xp_char_t* mode;

			if (epa->mode == SSE_AWK_IO_FILE_READ)
				mode = SSE_T("r");
			else if (epa->mode == SSE_AWK_IO_FILE_WRITE)
				mode = SSE_T("w");
			else if (epa->mode == SSE_AWK_IO_FILE_APPEND)
				mode = SSE_T("a");
			else return -1; /* TODO: any way to set the error number? */

xp_printf (SSE_TEXT("opending %s of type %d (file)\n"),  epa->name, epa->type);
			handle = fopen_t (epa->name, mode);
			if (handle == NULL) return -1;

			epa->handle = (void*)handle;
			return 1;
		}

		case SSE_AWK_IO_CLOSE:
		{
xp_printf (SSE_TEXT("closing %s of type %d (file)\n"),  epa->name, epa->type);
			fclose ((FILE*)epa->handle);
			epa->handle = NULL;
			return 0;
		}

		case SSE_AWK_IO_READ:
		{
			if (fgets_t (data, size, (FILE*)epa->handle) == SSE_NULL) 
				return 0;
			return sse_awk_strlen(data);
		}

		case SSE_AWK_IO_WRITE:
		{
			/* TODO: how to return error or 0 */
			fputs_t (data, /*size,*/ (FILE*)epa->handle);
			return size;
		}

		case SSE_AWK_IO_FLUSH:
		{
			if (fflush ((FILE*)epa->handle) == EOF) return -1;
			return 0;
		}

		case SSE_AWK_IO_NEXT:
		{
			return -1;
		}

	}

	return -1;
}

static int open_extio_console (sse_awk_extio_t* epa);
static int close_extio_console (sse_awk_extio_t* epa);
static int next_extio_console (sse_awk_extio_t* epa);

static xp_size_t infile_no = 0;
static const xp_char_t* infiles[10000] =
{
	SSE_T(""),
	SSE_NULL
};


static xp_ssize_t process_extio_console (
	int cmd, void* arg, xp_char_t* data, xp_size_t size)
{
	sse_awk_extio_t* epa = (sse_awk_extio_t*)arg;

	if (cmd == SSE_AWK_IO_OPEN)
	{
		return open_extio_console (epa);
	}
	else if (cmd == SSE_AWK_IO_CLOSE)
	{
		return close_extio_console (epa);
	}
	else if (cmd == SSE_AWK_IO_READ)
	{
		while (fgets_t (data, size, epa->handle) == SSE_NULL)
		{
			/* it has reached the end of the current file.
			 * open the next file if available */
			if (infiles[infile_no] == SSE_NULL) 
			{
				/* no more input console */

				/* is this correct??? */
				/*
				if (epa->handle != SSE_NULL &&
				    epa->handle != stdin &&
				    epa->handle != stdout &&
				    epa->handle != stderr) fclose (epa->handle);
				epa->handle = SSE_NULL;
				*/

				return 0;
			}

			if (infiles[infile_no][0] == SSE_T('\0'))
			{
				if (epa->handle != SSE_NULL &&
				    epa->handle != stdin &&
				    epa->handle != stdout &&
				    epa->handle != stderr) fclose (epa->handle);
				epa->handle = stdin;
			}
			else
			{
				FILE* fp = fopen_t (infiles[infile_no], SSE_T("r"));
				if (fp == SSE_NULL)
				{
xp_printf (SSE_TEXT("failed to open the next console of type %x - fopen failure\n"), epa->type);
					return -1;
				}

				if (epa->handle != SSE_NULL &&
				    epa->handle != stdin &&
				    epa->handle != stdout &&
				    epa->handle != stderr) fclose (epa->handle);

xp_printf (SSE_TEXT("open the next console [%s]\n"), infiles[infile_no]);
				epa->handle = fp;
			}

			infile_no++;	
		}

		return sse_awk_strlen(data);
	}
	else if (cmd == SSE_AWK_IO_WRITE)
	{
		/* TODO: how to return error or 0 */
		fputs_t (data, /*size,*/ (FILE*)epa->handle);
		/*MessageBox (NULL, data, data, MB_OK);*/
		return size;
	}
	else if (cmd == SSE_AWK_IO_FLUSH)
	{
		if (fflush ((FILE*)epa->handle) == EOF) return -1;
		return 0;
	}
	else if (cmd == SSE_AWK_IO_NEXT)
	{
		return next_extio_console (epa);
	}

	return -1;
}

static int open_extio_console (sse_awk_extio_t* epa)
{
	/* TODO: OpenConsole in GUI APPLICATION */

	/* epa->name is always empty for console */
	xp_assert (epa->name[0] == SSE_T('\0'));

xp_printf (SSE_TEXT("opening console[%s] of type %x\n"), epa->name, epa->type);

	if (epa->mode == SSE_AWK_IO_CONSOLE_READ)
	{
		if (infiles[infile_no] == SSE_NULL)
		{
			/* no more input file */
xp_printf (SSE_TEXT("console - no more file\n"));;
			return 0;
		}

		if (infiles[infile_no][0] == SSE_T('\0'))
		{
xp_printf (SSE_T("    console(r) - <standard input>\n"));
			epa->handle = stdin;
		}
		else
		{
			/* a temporary variable fp is used here not to change 
			 * any fields of epa when the open operation fails */
			FILE* fp = fopen_t (infiles[infile_no], SSE_T("r"));
			if (fp == SSE_NULL)
			{
xp_printf (SSE_TEXT("failed to open console of type %x - fopen failure\n"), epa->type);
				return -1;
			}

xp_printf (SSE_T("    console(r) - %s\n"), infiles[infile_no]);
			if (sse_awk_setconsolename (
				epa->run, infiles[infile_no], 
				sse_awk_strlen(infiles[infile_no])) == -1)
			{
				fclose (fp);
				return -1;
			}

			epa->handle = fp;
		}

		infile_no++;
		return 1;
	}
	else if (epa->mode == SSE_AWK_IO_CONSOLE_WRITE)
	{
xp_printf (SSE_T("    console(w) - <standard output>\n"));
		/* TODO: does output console has a name??? */
		/*sse_awk_setconsolename (SSE_T(""));*/
		epa->handle = stdout;
		return 1;
	}

	return -1;
}

static int close_extio_console (sse_awk_extio_t* epa)
{
xp_printf (SSE_TEXT("closing console of type %x\n"), epa->type);

	if (epa->handle != SSE_NULL &&
	    epa->handle != stdin && 
	    epa->handle != stdout && 
	    epa->handle != stderr)
	{
		fclose (epa->handle);
	}

	/* TODO: CloseConsole in GUI APPLICATION */
	return 0;
}

static int next_extio_console (sse_awk_extio_t* epa)
{
	int n;
	FILE* fp = epa->handle;
xp_printf (SSE_TEXT("switching console[%s] of type %x\n"), epa->name, epa->type);

	n = open_extio_console(epa);
	if (n == -1) return -1;

	if (n == 0) 
	{
		/* if there is no more file, keep the previous handle */
		return 0;
	}

	if (fp != SSE_NULL && fp != stdin && 
	    fp != stdout && fp != stderr) fclose (fp);

	return n;
}


sse_awk_t* app_awk = NULL;
void* app_run = NULL;

#ifdef _WIN32
static BOOL WINAPI __stop_run (DWORD ctrl_type)
{
	if (ctrl_type == CTRL_C_EVENT ||
	    ctrl_type == CTRL_CLOSE_EVENT)
	{
		sse_awk_stop (app_awk, app_run);
		return TRUE;
	}

	return FALSE;
}
#else
static void __stop_run (int sig)
{
	signal  (SIGINT, SIG_IGN);
	sse_awk_stop (app_awk, app_run);
	//sse_awk_stoprun (awk, handle);
	/*sse_awk_stopallruns (awk); */
	signal  (SIGINT, __stop_run);
}
#endif

static void __on_run_start (sse_awk_t* awk, void* handle, void* arg)
{
	app_awk = awk;	
	app_run = handle;
xp_printf (SSE_T("AWK PRORAM ABOUT TO START...\n"));
}

static void __on_run_end (sse_awk_t* awk, void* handle, int errnum, void* arg)
{
	if (errnum != SSE_AWK_ENOERR)
	{
		xp_printf (SSE_T("AWK PRORAM ABOUT TO END WITH AN ERROR - %d - %s\n"), errnum, sse_awk_geterrstr (errnum));
	}
	else xp_printf (SSE_T("AWK PRORAM ENDED SUCCESSFULLY\n"));

	app_awk = NULL;	
	app_run = NULL;
}

#ifdef _WIN32
typedef struct syscas_data_t syscas_data_t;
struct syscas_data_t
{
	HANDLE heap;
};
#endif

static void* __awk_malloc (xp_size_t n, void* custom_data)
{
#ifdef _WIN32
	return HeapAlloc (((syscas_data_t*)custom_data)->heap, 0, n);
#else
	return malloc (n);
#endif
}

static void* __awk_realloc (void* ptr, xp_size_t n, void* custom_data)
{
#ifdef _WIN32
	/* HeapReAlloc behaves differently from realloc */
	if (ptr == NULL)
		return HeapAlloc (((syscas_data_t*)custom_data)->heap, 0, n);
	else
		return HeapReAlloc (((syscas_data_t*)custom_data)->heap, 0, ptr, n);
#else
	return realloc (ptr, n);
#endif
}

static void __awk_free (void* ptr, void* custom_data)
{
#ifdef _WIN32
	HeapFree (((syscas_data_t*)custom_data)->heap, 0, ptr);
#else
	free (ptr);
#endif
}

#if defined(__STAND_ALONE) && !defined(_WIN32)
static int __main (int argc, char* argv[])
#else
static int __main (int argc, xp_char_t* argv[])
#endif
{
	sse_awk_t* awk;
	sse_awk_srcios_t srcios;
	sse_awk_runcbs_t runcbs;
	sse_awk_runios_t runios;
	sse_awk_runarg_t runarg[10];
	sse_awk_syscas_t syscas;
	struct src_io src_io = { NULL, NULL };
	int opt, i, file_count = 0;
#ifdef _WIN32
	syscas_data_t syscas_data;
#endif

	opt = SSE_AWK_EXPLICIT | SSE_AWK_UNIQUE | SSE_AWK_HASHSIGN |
		/*SSE_AWK_DBLSLASHES |*/
		SSE_AWK_SHADING | SSE_AWK_IMPLICIT | SSE_AWK_SHIFT | 
		SSE_AWK_EXTIO | SSE_AWK_BLOCKLESS | SSE_AWK_STRINDEXONE | 
		SSE_AWK_STRIPSPACES | SSE_AWK_NEWLINE;

	if (argc <= 1)
	{
		xp_printf (SSE_T("Usage: %s [-m] source_file [data_file ...]\n"), argv[0]);
		return -1;
	}

	for (i = 1; i < argc; i++)
	{
#if defined(__STAND_ALONE) && !defined(_WIN32)
		if (strcmp(argv[i], "-m") == 0)
#else
		if (sse_awk_strcmp(argv[i], SSE_T("-m")) == 0)
#endif
		{
			opt |= SSE_AWK_RUNMAIN;
		}
		else if (file_count == 0)
		{
			src_io.input_file = argv[i];
			file_count++;
		}
		else if (file_count >= 1 && file_count < xp_countof(infiles)-1)
		{
			infiles[file_count-1] = argv[i];
			infiles[file_count] = SSE_NULL;
			file_count++;
		}
		else
		{
			xp_printf (SSE_T("Usage: %s [-m] [-f source_file] [data_file ...]\n"), argv[0]);
			return -1;
		}
	}

	memset (&syscas, 0, sizeof(syscas));
	syscas.malloc = __awk_malloc;
	syscas.realloc = __awk_realloc;
	syscas.free = __awk_free;

	syscas.lock = NULL;
	syscas.unlock = NULL;

#ifdef SSE_CHAR_IS_MCHAR
	syscas.is_upper  = isupper;
	syscas.is_lower  = islower;
	syscas.is_alpha  = isalpha;
	syscas.is_digit  = isdigit;
	syscas.is_xdigit = isxdigit;
	syscas.is_alnum  = isalnum;
	syscas.is_space  = isspace;
	syscas.is_print  = isprint;
	syscas.is_graph  = isgraph;
	syscas.is_cntrl  = iscntrl;
	syscas.is_punct  = ispunct;
	syscas.to_upper  = toupper;
	syscas.to_lower  = tolower;
#else
	syscas.is_upper  = iswupper;
	syscas.is_lower  = iswlower;
	syscas.is_alpha  = iswalpha;
	syscas.is_digit  = iswdigit;
	syscas.is_xdigit = iswxdigit;
	syscas.is_alnum  = iswalnum;
	syscas.is_space  = iswspace;
	syscas.is_print  = iswprint;
	syscas.is_graph  = iswgraph;
	syscas.is_cntrl  = iswcntrl;
	syscas.is_punct  = iswpunct;
	syscas.to_upper  = towupper;
	syscas.to_lower  = towlower;
#endif
	syscas.memcpy = memcpy;
	syscas.memset = memset;
	syscas.sprintf = xp_sprintf;
	syscas.dprintf = __dprintf;
	syscas.abort = abort;

#ifdef _WIN32
	syscas_data.heap = HeapCreate (0, 1000000, 1000000);
	if (syscas_data.heap == NULL)
	{
		xp_printf (SSE_T("Error: cannot create an awk heap\n"));
		return -1;
	}

	syscas.custom_data = &syscas_data;
#endif

	if ((awk = sse_awk_open(&syscas)) == SSE_NULL) 
	{
#ifdef _WIN32
		HeapDestroy (syscas_data.heap);
#endif
		xp_printf (SSE_T("Error: cannot open awk\n"));
		return -1;
	}

	sse_awk_setopt (awk, opt);

	srcios.in = process_source;
	srcios.out = dump_source;
	srcios.custom_data = &src_io;

	if (sse_awk_parse (awk, &srcios) == -1) 
	{
		int errnum = sse_awk_geterrnum(awk);
#if defined(__STAND_ALONE) && !defined(_WIN32) && defined(SSE_CHAR_IS_WCHAR)
		xp_printf (
			SSE_T("ERROR: cannot parse program - line %u [%d] %ls\n"), 
			(unsigned int)sse_awk_getsrcline(awk), 
			errnum, sse_awk_geterrstr(errnum));
#else
		xp_printf (
			SSE_T("ERROR: cannot parse program - line %u [%d] %s\n"), 
			(unsigned int)sse_awk_getsrcline(awk), 
			errnum, sse_awk_geterrstr(errnum));
#endif
		sse_awk_close (awk);
		return -1;
	}

#ifdef _WIN32
	SetConsoleCtrlHandler (__stop_run, TRUE);
#else
	signal (SIGINT, __stop_run);
#endif

	runios.pipe = process_extio_pipe;
	runios.coproc = SSE_NULL;
	runios.file = process_extio_file;
	runios.console = process_extio_console;

	runcbs.on_start = __on_run_start;
	runcbs.on_end = __on_run_end;
	runcbs.custom_data = SSE_NULL;

	runarg[0].ptr = SSE_T("argument 0");
	runarg[0].len = sse_awk_strlen(runarg[0].ptr);
	runarg[1].ptr = SSE_T("argumetn 1");
	runarg[1].len = sse_awk_strlen(runarg[1].ptr);
	runarg[2].ptr = SSE_T("argumetn 2");
	runarg[2].len = sse_awk_strlen(runarg[2].ptr);
	runarg[3].ptr = SSE_NULL;
	runarg[3].len = 0;

	if (sse_awk_run (awk, &runios, &runcbs, runarg) == -1)
	{
		int errnum = sse_awk_geterrnum(awk);
#if defined(__STAND_ALONE) && !defined(_WIN32) && defined(SSE_CHAR_IS_WCHAR)
		xp_printf (
			SSE_T("error: cannot run program - [%d] %ls\n"), 
			errnum, sse_awk_geterrstr(errnum));
#else
		xp_printf (
			SSE_T("error: cannot run program - [%d] %s\n"), 
			errnum, sse_awk_geterrstr(errnum));
#endif

		sse_awk_close (awk);
		return -1;
	}

	sse_awk_close (awk);
#ifdef _WIN32
	HeapDestroy (syscas_data.heap);
#endif
	return 0;
}

#ifdef _WIN32
/*
NTSYSAPI PTEB NTAPI NtCurrentTeb();
Function NtCurrentTeb returns address of TEB (Thread Environment Block) for calling thread. 
NtCurrentTeb isn't typical NT CALL realised via INT 2E, becouse TEB is accessable at address fs:[0018h].
Microsoft declare NtCurrentTeb as __cdecl, but ntdll.dll export it as __stdcall (it don't have metter, becouse function don't have any parameters), so you cannot use ntdll.dll export. In this case the better way is write NtCurrentTeb manually, declaring it as __cdecl. 

typedef UCHAR BOOLEAN;

typedef struct _TEB {

  NT_TIB                  Tib;
  PVOID                   EnvironmentPointer;
  CLIENT_ID               Cid;
  PVOID                   ActiveRpcInfo;
  PVOID                   ThreadLocalStoragePointer;
  PPEB                    Peb;
  ULONG                   LastErrorValue;
  ULONG                   CountOfOwnedCriticalSections;
  PVOID                   CsrClientThread;
  PVOID                   Win32ThreadInfo;
  ULONG                   Win32ClientInfo[0x1F];
  PVOID                   WOW32Reserved;
  ULONG                   CurrentLocale;
  ULONG                   FpSoftwareStatusRegister;
  PVOID                   SystemReserved1[0x36];
  PVOID                   Spare1;
  ULONG                   ExceptionCode;
  ULONG                   SpareBytes1[0x28];
  PVOID                   SystemReserved2[0xA];
  ULONG                   GdiRgn;
  ULONG                   GdiPen;
  ULONG                   GdiBrush;
  CLIENT_ID               RealClientId;
  PVOID                   GdiCachedProcessHandle;
  ULONG                   GdiClientPID;
  ULONG                   GdiClientTID;
  PVOID                   GdiThreadLocaleInfo;
  PVOID                   UserReserved[5];
  PVOID                   GlDispatchTable[0x118];
  ULONG                   GlReserved1[0x1A];
  PVOID                   GlReserved2;
  PVOID                   GlSectionInfo;
  PVOID                   GlSection;
  PVOID                   GlTable;
  PVOID                   GlCurrentRC;
  PVOID                   GlContext;
  NTSTATUS                LastStatusValue;
  UNICODE_STRING          StaticUnicodeString;
  WCHAR                   StaticUnicodeBuffer[0x105];
  PVOID                   DeallocationStack;
  PVOID                   TlsSlots[0x40];
  LIST_ENTRY              TlsLinks;
  PVOID                   Vdm;
  PVOID                   ReservedForNtRpc;
  PVOID                   DbgSsReserved[0x2];
  ULONG                   HardErrorDisabled;
  PVOID                   Instrumentation[0x10];
  PVOID                   WinSockData;
  ULONG                   GdiBatchCount;
  ULONG                   Spare2;
  ULONG                   Spare3;
  ULONG                   Spare4;
  PVOID                   ReservedForOle;
  ULONG                   WaitingOnLoaderLock;
  PVOID                   StackCommit;
  PVOID                   StackCommitMax;
  PVOID                   StackReserved;

} TEB, *PTEB;

typedef struct _PEB {
  BOOLEAN                 InheritedAddressSpace;
  BOOLEAN                 ReadImageFileExecOptions;
  BOOLEAN                 BeingDebugged;
  BOOLEAN                 Spare;
  HANDLE                  Mutant;
  PVOID                   ImageBaseAddress;
  PPEB_LDR_DATA           LoaderData;
  PRTL_USER_PROCESS_PARAMETERS ProcessParameters;
  PVOID                   SubSystemData;
  PVOID                   ProcessHeap;
  PVOID                   FastPebLock;
  PPEBLOCKROUTINE         FastPebLockRoutine;
  PPEBLOCKROUTINE         FastPebUnlockRoutine;
  ULONG                   EnvironmentUpdateCount;
  PPVOID                  KernelCallbackTable;
  PVOID                   EventLogSection;
  PVOID                   EventLog;
  PPEB_FREE_BLOCK         FreeList;
  ULONG                   TlsExpansionCounter;
  PVOID                   TlsBitmap;
  ULONG                   TlsBitmapBits[0x2];
  PVOID                   ReadOnlySharedMemoryBase;
  PVOID                   ReadOnlySharedMemoryHeap;
  PPVOID                  ReadOnlyStaticServerData;
  PVOID                   AnsiCodePageData;
  PVOID                   OemCodePageData;
  PVOID                   UnicodeCaseTableData;
  ULONG                   NumberOfProcessors;
  ULONG                   NtGlobalFlag;
  BYTE                    Spare2[0x4];
  LARGE_INTEGER           CriticalSectionTimeout;
  ULONG                   HeapSegmentReserve;
  ULONG                   HeapSegmentCommit;
  ULONG                   HeapDeCommitTotalFreeThreshold;
  ULONG                   HeapDeCommitFreeBlockThreshold;
  ULONG                   NumberOfHeaps;
  ULONG                   MaximumNumberOfHeaps;
  PPVOID                  *ProcessHeaps;
  PVOID                   GdiSharedHandleTable;
  PVOID                   ProcessStarterHelper;
  PVOID                   GdiDCAttributeList;
  PVOID                   LoaderLock;
  ULONG                   OSMajorVersion;
  ULONG                   OSMinorVersion;
  ULONG                   OSBuildNumber;
  ULONG                   OSPlatformId;
  ULONG                   ImageSubSystem;
  ULONG                   ImageSubSystemMajorVersion;
  ULONG                   ImageSubSystemMinorVersion;
  ULONG                   GdiHandleBuffer[0x22];
  ULONG                   PostProcessInitRoutine;
  ULONG                   TlsExpansionBitmap;
  BYTE                    TlsExpansionBitmapBits[0x80];
  ULONG                   SessionId;

} PEB, *PPEB;


*/
void* get_current_teb ()
{
	_asm 
	{
		mov eax, fs:[0x18]
	}
}

void* get_current_peb ()
{
	void* teb = get_current_teb ();
	return *(void**)((char*)teb + 0x30);
}

int is_debugger_present ()
{
	void *peb = get_current_peb ();
	return *((char*)peb+0x02);
}


int is_debugger_present2 ()
{
	_asm
	{
		mov eax, fs:[0x18]
		mov ebx, [eax+0x30]
		xor eax, eax
		mov al, byte ptr[ebx+0x02]
	}
}
#endif

#if defined(__STAND_ALONE) && !defined(_WIN32)
int main (int argc, char* argv[])
#else
int xp_main (int argc, xp_char_t* argv[])
#endif
{
	int n;
#if defined(__linux) && defined(_DEBUG)
	mtrace ();
#endif
/*#if defined(_WIN32) && defined(_MSC_VER) && defined(_DEBUG)
	_CrtSetDbgFlag (_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF | _CRTDBG_CHECK_ALWAYS_DF);
#endif*/

#ifdef _WIN32
{
xp_char_t buf[xp_sizeof(xp_long_t)*8+2+2];
xp_size_t n;
n = sse_awk_longtostr (-0x7FFFFFFFFFFFFFFFi64, 16, SSE_T("0x"), buf, xp_countof(buf));
if (n == (xp_size_t)-1)
{
	xp_printf (SSE_T("cannot convert...\n"));
}
else xp_printf (SSE_T("%d, %s\n"), n, buf);
}

	if (IsDebuggerPresent ())
	{
		xp_printf (SSE_T("Running application in a debugger....\n"));
	}
	if (is_debugger_present ())
	{
		xp_printf (SSE_T("Running application in a debugger by is_debugger_present...\n"));
	}
	if (is_debugger_present2 ())
	{
		xp_printf (SSE_T("Running application in a debugger by is_debugger_present2...\n"));
	}
#endif

	n = __main (argc, argv);

#if defined(__linux) && defined(_DEBUG)
	muntrace ();
#endif
#if defined(_WIN32) && defined(_MSC_VER) && defined(_DEBUG)
	_CrtDumpMemoryLeaks ();
	wprintf (L"Press ENTER to quit\n");
	getchar ();
#endif

	return n;
}

